# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/pb/inc_qua.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/pb/inc_qua.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from qm.grpc.qm.pb import inc_qua_config_pb2 as qm_dot_pb_dot_inc__qua__config__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13qm/pb/inc_qua.proto\x12\x0bqm.grpc.qua\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1aqm/pb/inc_qua_config.proto\x1a\x19google/protobuf/any.proto\"\xb4\x94\x01\n\nQuaProgram\x12-\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x12\x38\n\tdynConfig\x18\x02 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.DynamicConfig\x12.\n\x06script\x18\x05 \x01(\x0b\x32\x1e.qm.grpc.qua.QuaProgram.Script\x12@\n\x0f\x63ompilerOptions\x18\x06 \x01(\x0b\x32\'.qm.grpc.qua.QuaProgram.CompilerOptions\x12\x36\n\x0eresultAnalysis\x18\x07 \x01(\x0b\x32\x1e.qm.grpc.qua.QuaResultAnalysis\x12\x34\n\rconfig_update\x18\x08 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x1a\xd2\x01\n\x0f\x43ompilerOptions\x12.\n\"useExperimentalCalculationCompiler\x18\x01 \x01(\x08\x42\x02\x18\x01\x12&\n\x1aoptimizeMergeCodeExecution\x18\x02 \x01(\x08\x42\x02\x18\x01\x12%\n\x19optimizeWriteReadCommands\x18\x03 \x01(\x08\x42\x02\x18\x01\x12\x1d\n\x11skipOptimizations\x18\x04 \x03(\tB\x02\x18\x01\x12\x12\n\x06strict\x18\x05 \x01(\x08\x42\x02\x18\x01\x12\r\n\x05\x66lags\x18\x06 \x03(\t\x1a\xcb\x01\n\x06Script\x12\x39\n\tvariables\x18\x01 \x03(\x0b\x32&.qm.grpc.qua.QuaProgram.VarDeclaration\x12J\n\x0f\x65xternalStreams\x18\x02 \x03(\x0b\x32\x31.qm.grpc.qua.QuaProgram.ExternalStreamDeclaration\x12:\n\x04\x62ody\x18\n \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x1a\xb1\x02\n\x0eVarDeclaration\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.qm.grpc.qua.QuaProgram.Type\x12\x0c\n\x04size\x18\x03 \x01(\x05\x12\x38\n\x05value\x18\x04 \x03(\x0b\x32).qm.grpc.qua.QuaProgram.LiteralExpression\x12\x0b\n\x03\x64im\x18\x05 \x01(\x05\x12\x15\n\risInputStream\x18\x06 \x01(\x08\x12I\n\x0cstructMember\x18\x07 \x01(\x0b\x32\x33.qm.grpc.qua.QuaProgram.VarDeclaration.StructMember\x1a.\n\x0cStructMember\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08position\x18\x02 \x01(\x05\x1a\xa3\x01\n\x19\x45xternalStreamDeclaration\x12\x11\n\tstream_id\x18\x01 \x01(\r\x12=\n\rexpectedTypes\x18\x02 \x03(\x0b\x32&.qm.grpc.qua.QuaProgram.VarDeclaration\x12\x34\n\tdirection\x18\x03 \x01(\x0e\x32!.qm.grpc.qua.QuaProgram.Direction\x1a\x34\n\x17QuantumElementReference\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a+\n\x0ePulseReference\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x37\n\x1aIntegrationWeightReference\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\xbf\x11\n\x0c\x41nyStatement\x12\x35\n\x04play\x18\x01 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.PlayStatementH\x00\x12;\n\x07measure\x18\x02 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.MeasureStatementH\x00\x12\x35\n\x04wait\x18\x03 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.WaitStatementH\x00\x12\x35\n\x04sync\x18\x04 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.SyncStatementH\x00\x12\x31\n\x02if\x18\x05 \x01(\x0b\x32#.qm.grpc.qua.QuaProgram.IfStatementH\x00\x12=\n\x06\x61ssign\x18\x06 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AssignmentStatementH\x00\x12\x33\n\x03\x66or\x18\x07 \x01(\x0b\x32$.qm.grpc.qua.QuaProgram.ForStatementH\x00\x12\x33\n\x03rus\x18\x08 \x01(\x0b\x32$.qm.grpc.qua.QuaProgram.RusStatementH\x00\x12\x37\n\x05\x61lign\x18\t \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.AlignStatementH\x00\x12K\n\x0fupdateFrequency\x18\n \x01(\x0b\x32\x30.qm.grpc.qua.QuaProgram.UpdateFrequencyStatementH\x00\x12?\n\tzRotation\x18\x0b \x01(\x0b\x32*.qm.grpc.qua.QuaProgram.ZRotationStatementH\x00\x12\x37\n\x05pause\x18\x0c \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.PauseStatementH\x00\x12\x35\n\x04save\x18\r \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.SaveStatementH\x00\x12;\n\x07\x66orEach\x18\x0e \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.ForEachStatementH\x00\x12I\n\x0ewaitForTrigger\x18\x0f \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.WaitForTriggerStatementH\x00\x12\x39\n\x04qrun\x18\x10 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.QRunStatementB\x02\x18\x01H\x00\x12M\n\x10updateCorrection\x18\x11 \x01(\x0b\x32\x31.qm.grpc.qua.QuaProgram.UpdateCorrectionStatementH\x00\x12\x41\n\nresetPhase\x18\x12 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.ResetPhaseStatementH\x00\x12\x41\n\nrampToZero\x18\x13 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.RampToZeroStatementH\x00\x12\x41\n\nresetFrame\x18\x14 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.ResetFrameStatementH\x00\x12\x43\n\x0bsetDcOffset\x18\x15 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.SetDcOffsetStatementH\x00\x12Q\n\x12\x61\x64vanceInputStream\x18\x16 \x01(\x0b\x32\x33.qm.grpc.qua.QuaProgram.AdvanceInputStreamStatementH\x00\x12\x45\n\x0cstrictTiming\x18\x17 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.StrictTimingStatementH\x00\x12O\n\x11\x66\x61stFrameRotation\x18\x18 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.FastFrameRotationStatementH\x00\x12_\n\x19updateOscillatorFrequency\x18\x19 \x01(\x0b\x32:.qm.grpc.qua.QuaProgram.UpdateOscillatorFrequencyStatementH\x00\x12M\n\x10resetGlobalPhase\x18\x1b \x01(\x0b\x32\x31.qm.grpc.qua.QuaProgram.ResetGlobalPhaseStatementH\x00\x12\x45\n\x0cloadWaveform\x18\x1c \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.LoadWaveformStatementH\x00\x12U\n\x14sendToExternalStream\x18\x1d \x01(\x0b\x32\x35.qm.grpc.qua.QuaProgram.SendToExternalStreamStatementH\x00\x12_\n\x19receiveFromExternalStream\x18\x1e \x01(\x0b\x32:.qm.grpc.qua.QuaProgram.ReceiveFromExternalStreamStatementH\x00\x12]\n\x18globalVariableAssignment\x18\x1f \x01(\x0b\x32\x39.qm.grpc.qua.QuaProgram.GlobalVariableAssignmentStatementH\x00\x12?\n\tarbitrary\x18\x64 \x01(\x0b\x32*.qm.grpc.qua.QuaProgram.ArbitraryStatementH\x00\x12M\n\x10\x61rbitraryContext\x18\x65 \x01(\x0b\x32\x31.qm.grpc.qua.QuaProgram.ArbitraryContextStatementH\x00\x42\x11\n\x0fstatement_oneof\x1a\x80\x02\n\rAmpMultiplier\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x37\n\x02v0\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x37\n\x02v1\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x37\n\x02v2\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x37\n\x02v3\x18\x04 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1a\x8f\x01\n\x13RampToZeroStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12.\n\x08\x64uration\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x1aT\n\tRampPulse\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12:\n\x05value\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1a\xe4\x02\n\x05\x43hirp\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x41\n\nscalarRate\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpressionH\x00\x12\x42\n\tarrayRate\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpressionH\x00\x12\x32\n\x05units\x18\x03 \x01(\x0e\x32#.qm.grpc.qua.QuaProgram.Chirp.Units\x12\r\n\x05times\x18\x04 \x03(\r\x12\x15\n\rcontinueChirp\x18\x05 \x01(\x08\"e\n\x05Units\x12\x10\n\x0cHzPerNanoSec\x10\x00\x12\x11\n\rmHzPerNanoSec\x10\x01\x12\x11\n\ruHzPerNanoSec\x10\x02\x12\x11\n\rnHzPerNanoSec\x10\x03\x12\x11\n\rpHzPerNanoSec\x10\x04\x42\x06\n\x04rate\x1a\xa9\x05\n\rPlayStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x39\n\x05pulse\x18\x02 \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.PulseReferenceB\x02\x18\x01\x12<\n\nnamedPulse\x18\x07 \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.PulseReferenceH\x00\x12\x36\n\trampPulse\x18\x08 \x01(\x0b\x32!.qm.grpc.qua.QuaProgram.RampPulseH\x00\x12\x32\n\x03\x61mp\x18\x03 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.AmpMultiplier\x12=\n\x08\x64uration\x18\x04 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12>\n\tcondition\x18\x05 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x43\n\x0eport_condition\x18\x0c \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x13\n\x0btargetInput\x18\x06 \x01(\t\x12,\n\x05\x63hirp\x18\t \x01(\x0b\x32\x1d.qm.grpc.qua.QuaProgram.Chirp\x12=\n\x08truncate\x18\n \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x16\n\x0etimestampLabel\x18\x0b \x01(\tB\x0b\n\tpulseType\x1a\xaf\x02\n\x18UpdateFrequencyStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12:\n\x05value\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x45\n\x05units\x18\x03 \x01(\x0e\x32\x36.qm.grpc.qua.QuaProgram.UpdateFrequencyStatement.Units\x12\x11\n\tkeepPhase\x18\x04 \x01(\x08\"3\n\x05Units\x12\x06\n\x02Hz\x10\x00\x12\x07\n\x03mHz\x10\x01\x12\x07\n\x03uHz\x10\x02\x12\x07\n\x03nHz\x10\x03\x12\x07\n\x03pHz\x10\x04\x1a\xf2\x03\n\"UpdateOscillatorFrequencyStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12:\n\x05value\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12O\n\x05units\x18\x03 \x01(\x0e\x32@.qm.grpc.qua.QuaProgram.UpdateOscillatorFrequencyStatement.Units\x12\x11\n\tkeepPhase\x18\x04 \x01(\x08\x12l\n\x0fupdateComponent\x18\x05 \x01(\x0e\x32S.qm.grpc.qua.QuaProgram.UpdateOscillatorFrequencyStatement.UpdateComponentSelection\"*\n\x05Units\x12\x07\n\x03GHz\x10\x00\x12\x07\n\x03MHz\x10\x01\x12\x07\n\x03KHz\x10\x02\x12\x06\n\x02Hz\x10\x03\"H\n\x18UpdateComponentSelection\x12\x08\n\x04\x62oth\x10\x00\x12\x0f\n\x0bupconverter\x10\x01\x12\x11\n\rdownconverter\x10\x02\x1a\xb7\x01\n\x14SetDcOffsetStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x18\n\x10qeInputReference\x18\x02 \x01(\t\x12;\n\x06offset\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1a\x9a\x01\n\x12ZRotationStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x03 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12:\n\x05value\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1a_\n\x13ResetFrameStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x02 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x1a\xde\x01\n\x1a\x46\x61stFrameRotationStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12;\n\x06\x63osine\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x39\n\x04sine\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1aZ\n\x0e\x41lignStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x03(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x1a\x9d\x01\n\x0eMeasureProcess\x12>\n\x06\x61nalog\x18\x01 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.AnalogMeasureProcessH\x00\x12@\n\x07\x64igital\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.DigitalMeasureProcessH\x00\x42\t\n\x07process\x1a\xe4\x0e\n\x14\x41nalogMeasureProcess\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12W\n\x0f\x62\x61reIntegration\x18\x01 \x01(\x0b\x32<.qm.grpc.qua.QuaProgram.AnalogMeasureProcess.BareIntegrationH\x00\x12Y\n\x10\x64\x65modIntegration\x18\x02 \x01(\x0b\x32=.qm.grpc.qua.QuaProgram.AnalogMeasureProcess.DemodIntegrationH\x00\x12U\n\x0erawTimeTagging\x18\x03 \x01(\x0b\x32;.qm.grpc.qua.QuaProgram.AnalogMeasureProcess.RawTimeTaggingH\x00\x12_\n\x13\x64ualBareIntegration\x18\x04 \x01(\x0b\x32@.qm.grpc.qua.QuaProgram.AnalogMeasureProcess.DualBareIntegrationH\x00\x12\x61\n\x14\x64ualDemodIntegration\x18\x05 \x01(\x0b\x32\x41.qm.grpc.qua.QuaProgram.AnalogMeasureProcess.DualDemodIntegrationH\x00\x12]\n\x12highResTimeTagging\x18\x06 \x01(\x0b\x32?.qm.grpc.qua.QuaProgram.AnalogMeasureProcess.HighResTimeTaggingH\x00\x1a\xae\x01\n\x0f\x42\x61reIntegration\x12G\n\x0bintegration\x18\x01 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12;\n\x06target\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnalogProcessTarget\x12\x15\n\relementOutput\x18\t \x01(\t\x1a\x96\x02\n\x13\x44ualBareIntegration\x12H\n\x0cintegration1\x18\x01 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12H\n\x0cintegration2\x18\x02 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12;\n\x06target\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnalogProcessTarget\x12\x16\n\x0e\x65lementOutput1\x18\t \x01(\t\x12\x16\n\x0e\x65lementOutput2\x18\n \x01(\t\x1a\xaf\x01\n\x10\x44\x65modIntegration\x12G\n\x0bintegration\x18\x01 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12;\n\x06target\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnalogProcessTarget\x12\x15\n\relementOutput\x18\t \x01(\t\x1a\x97\x02\n\x14\x44ualDemodIntegration\x12H\n\x0cintegration1\x18\x01 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12H\n\x0cintegration2\x18\x02 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12;\n\x06target\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnalogProcessTarget\x12\x16\n\x0e\x65lementOutput1\x18\t \x01(\t\x12\x16\n\x0e\x65lementOutput2\x18\n \x01(\t\x1a\xb4\x01\n\x0eRawTimeTagging\x12=\n\x06target\x18\x01 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x12;\n\ttargetLen\x18\x02 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x12\x0f\n\x07maxTime\x18\x03 \x01(\x05\x12\x15\n\relementOutput\x18\t \x01(\t\x1a\xb8\x01\n\x12HighResTimeTagging\x12=\n\x06target\x18\x01 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x12;\n\ttargetLen\x18\x02 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x12\x0f\n\x07maxTime\x18\x03 \x01(\x05\x12\x15\n\relementOutput\x18\t \x01(\tB\t\n\x07process\x1a\xf9\x03\n\x15\x44igitalMeasureProcess\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12V\n\x0erawTimeTagging\x18\x01 \x01(\x0b\x32<.qm.grpc.qua.QuaProgram.DigitalMeasureProcess.RawTimeTaggingH\x00\x12J\n\x08\x63ounting\x18\x02 \x01(\x0b\x32\x36.qm.grpc.qua.QuaProgram.DigitalMeasureProcess.CountingH\x00\x1a\xb4\x01\n\x0eRawTimeTagging\x12\x15\n\relementOutput\x18\x01 \x01(\t\x12=\n\x06target\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x12;\n\ttargetLen\x18\x03 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x12\x0f\n\x07maxTime\x18\x04 \x01(\x05\x1am\n\x08\x43ounting\x12\x16\n\x0e\x65lementOutputs\x18\x01 \x03(\t\x12\x38\n\x06target\x18\x02 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x12\x0f\n\x07maxTime\x18\x03 \x01(\x05\x42\t\n\x07process\x1a\xb8\x06\n\x13\x41nalogProcessTarget\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12X\n\rscalarProcess\x18\x01 \x01(\x0b\x32?.qm.grpc.qua.QuaProgram.AnalogProcessTarget.ScalarProcessTargetH\x00\x12X\n\rvectorProcess\x18\x02 \x01(\x0b\x32?.qm.grpc.qua.QuaProgram.AnalogProcessTarget.VectorProcessTargetH\x00\x1a\xa2\x01\n\x13ScalarProcessTarget\x12<\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionH\x00\x12\x43\n\tarrayCell\x18\x02 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.ArrayCellRefExpressionH\x00\x42\x08\n\x06target\x1a\xa3\x01\n\x13VectorProcessTarget\x12<\n\x05\x61rray\x18\x01 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x12N\n\x0ctimeDivision\x18\x02 \x01(\x0b\x32\x38.qm.grpc.qua.QuaProgram.AnalogProcessTarget.TimeDivision\x1a\x83\x02\n\x0cTimeDivision\x12\x43\n\x06sliced\x18\x02 \x01(\x0b\x32\x31.qm.grpc.qua.QuaProgram.AnalogTimeDivision.SlicedH\x00\x12M\n\x0b\x61\x63\x63umulated\x18\x03 \x01(\x0b\x32\x36.qm.grpc.qua.QuaProgram.AnalogTimeDivision.AccumulatedH\x00\x12O\n\x0cmovingWindow\x18\x04 \x01(\x0b\x32\x37.qm.grpc.qua.QuaProgram.AnalogTimeDivision.MovingWindowH\x00\x42\x0e\n\x0ctimeDivisionB\x0f\n\rprocessTarget\x1a\xc8\x01\n\x12\x41nalogTimeDivision\x1a.\n\x06Sliced\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x17\n\x0fsamplesPerChunk\x18\x01 \x01(\x05\x1a\x33\n\x0b\x41\x63\x63umulated\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x17\n\x0fsamplesPerChunk\x18\x01 \x01(\x05\x1aM\n\x0cMovingWindow\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x17\n\x0fsamplesPerChunk\x18\x01 \x01(\x05\x12\x17\n\x0f\x63hunksPerWindow\x18\x02 \x01(\x05\x1a\x8b\x03\n\rMeasureOutput\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12G\n\x0bintegration\x18\x01 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.IntegrationWeightReference\x12>\n\x08variable\x18\x02 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionB\x02\x18\x01\x12\x0e\n\x06output\x18\x03 \x01(\t\x12<\n\x06target\x18\x04 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.MeasureOutput.Target\x1a\x95\x01\n\x06Target\x12<\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionH\x00\x12\x43\n\tarrayCell\x18\x02 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.ArrayCellRefExpressionH\x00\x42\x08\n\x06target\x1a\xef\x02\n\x10MeasureStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x35\n\x05pulse\x18\x02 \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.PulseReference\x12\x32\n\x03\x61mp\x18\x03 \x01(\x0b\x32%.qm.grpc.qua.QuaProgram.AmpMultiplier\x12:\n\x07outputs\x18\x04 \x03(\x0b\x32%.qm.grpc.qua.QuaProgram.MeasureOutputB\x02\x18\x01\x12\x10\n\x08streamAs\x18\x05 \x01(\t\x12@\n\x10measureProcesses\x18\x07 \x03(\x0b\x32&.qm.grpc.qua.QuaProgram.MeasureProcess\x12\x16\n\x0etimestampLabel\x18\x08 \x01(\t\x1a\x94\x01\n\rWaitStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x03(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x39\n\x04time\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1aS\n\x12\x41rbitraryStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\"\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x1a\xa4\x01\n\x19\x41rbitraryContextStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\'\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x14.google.protobuf.AnyH\x00\x88\x01\x01\x12:\n\x04\x62ody\x18\x03 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollectionB\x07\n\x05_data\x1a\xa6\x03\n\x17WaitForTriggerStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x03(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12;\n\x0bpulseToPlay\x18\x02 \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.PulseReference\x12/\n\rglobalTrigger\x18\x03 \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00\x12V\n\relementOutput\x18\x04 \x01(\x0b\x32=.qm.grpc.qua.QuaProgram.WaitForTriggerStatement.ElementOutputH\x00\x12?\n\rtimeTagTarget\x18\x05 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x1a\x30\n\rElementOutput\x12\x0f\n\x07\x65lement\x18\x01 \x01(\t\x12\x0e\n\x06output\x18\x02 \x01(\tB\x08\n\x06source\x1a\x9d\x01\n\x19UpdateCorrectionStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x36\n\ncorrection\x18\x02 \x01(\x0b\x32\".qm.grpc.qua.QuaProgram.Correction\x1a_\n\x13ResetPhaseStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x1a(\n\x19ResetGlobalPhaseStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\xb1\x01\n\x1dSendToExternalStreamStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x43\n\x06stream\x18\x01 \x01(\x0b\x32\x33.qm.grpc.qua.QuaProgram.ExternalStreamRefExpression\x12>\n\x06struct\x18\x02 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.StructVarRefExpression\x1a\xb6\x01\n\"ReceiveFromExternalStreamStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x43\n\x06stream\x18\x01 \x01(\x0b\x32\x33.qm.grpc.qua.QuaProgram.ExternalStreamRefExpression\x12>\n\x06struct\x18\x02 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.StructVarRefExpression\x1am\n!GlobalVariableAssignmentStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\tvariables\x18\x01 \x03(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x1a\xfd\x01\n\nCorrection\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x37\n\x02\x63\x30\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x37\n\x02\x63\x31\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x37\n\x02\x63\x32\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x37\n\x02\x63\x33\x18\x04 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1a\x1c\n\rSyncStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x91\x01\n\x06\x45lseIf\x12>\n\tcondition\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12:\n\x04\x62ody\x18\x02 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x93\x02\n\x0bIfStatement\x12>\n\tcondition\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12:\n\x04\x62ody\x18\x02 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12:\n\x04\x65lse\x18\x03 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12/\n\x07\x65lseifs\x18\x04 \x03(\x0b\x32\x1e.qm.grpc.qua.QuaProgram.ElseIf\x12\x0e\n\x06unsafe\x18\x05 \x01(\x08\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x91\x02\n\x0c\x46orStatement\x12:\n\x04init\x18\x01 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12>\n\tcondition\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12<\n\x06update\x18\x03 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12:\n\x04\x62ody\x18\x04 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1aX\n\rQRunStatement\x12:\n\x04\x62ody\x18\x01 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a`\n\x15StrictTimingStatement\x12:\n\x04\x62ody\x18\x01 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\xbb\x02\n\x10\x46orEachStatement\x12M\n\x08iterator\x18\x01 \x03(\x0b\x32;.qm.grpc.qua.QuaProgram.ForEachStatement.VariableWithValues\x12:\n\x04\x62ody\x18\x02 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x8e\x01\n\x12VariableWithValues\x12:\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpression\x12<\n\x05\x61rray\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x1a\xd7\x01\n\x0cRusStatement\x12=\n\x07\x64oBlock\x18\x01 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12?\n\tfailBlock\x18\x02 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.StatementsCollection\x12:\n\x05until\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x91\x03\n\x13\x41ssignmentStatement\x12>\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionB\x02\x18\x01\x12?\n\nexpression\x18\x02 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x10\n\x04save\x18\x03 \x01(\x08\x42\x02\x18\x01\x12\x42\n\x06target\x18\x04 \x01(\x0b\x32\x32.qm.grpc.qua.QuaProgram.AssignmentStatement.Target\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x95\x01\n\x06Target\x12<\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionH\x00\x12\x43\n\tarrayCell\x18\x02 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.ArrayCellRefExpressionH\x00\x42\x08\n\x06target\x1a[\n\x0ePauseStatement\x12<\n\x03qes\x18\x01 \x03(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\xfd\x02\n\rSaveStatement\x12>\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionB\x02\x18\x01\x12\x0b\n\x03tag\x18\x02 \x01(\t\x12<\n\x06source\x18\x03 \x01(\x0b\x32,.qm.grpc.qua.QuaProgram.SaveStatement.Source\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\xd3\x01\n\x06Source\x12<\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionH\x00\x12\x43\n\tarrayCell\x18\x02 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.ArrayCellRefExpressionH\x00\x12<\n\x07literal\x18\x03 \x01(\x0b\x32).qm.grpc.qua.QuaProgram.LiteralExpressionH\x00\x42\x08\n\x06source\x1a\xdd\x01\n\x15LoadWaveformStatement\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12;\n\x02qe\x18\x01 \x01(\x0b\x32/.qm.grpc.qua.QuaProgram.QuantumElementReference\x12\x35\n\x05pulse\x18\x02 \x01(\x0b\x32&.qm.grpc.qua.QuaProgram.PulseReference\x12\x43\n\x0ewaveform_index\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x1aP\n\x14StatementsCollection\x12\x38\n\nstatements\x18\x01 \x03(\x0b\x32$.qm.grpc.qua.QuaProgram.AnyStatement\x1a\x87\x05\n\x13\x41nyScalarExpression\x12<\n\x08variable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionH\x00\x12<\n\x07literal\x18\x02 \x01(\x0b\x32).qm.grpc.qua.QuaProgram.LiteralExpressionH\x00\x12\x43\n\x0f\x62inaryOperation\x18\x03 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.BinaryExpressionH\x00\x12\x43\n\tarrayCell\x18\x04 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.ArrayCellRefExpressionH\x00\x12\x44\n\x0b\x61rrayLength\x18\x05 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayLengthExpressionH\x00\x12\x44\n\x0blibFunction\x18\x06 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.LibFunctionExpressionH\x00\x12>\n\x08\x66unction\x18\x07 \x01(\x0b\x32*.qm.grpc.qua.QuaProgram.FunctionExpressionH\x00\x12@\n\tbroadcast\x18\x08 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.BroadcastExpressionH\x00\x12H\n\x0eglobalVariable\x18\t \x01(\x0b\x32..qm.grpc.qua.QuaProgram.GlobalVarRefExpressionH\x00\x42\x12\n\x10\x65xpression_oneof\x1a\xc4\x01\n\x1b\x41\x64vanceInputStreamStatement\x12\x42\n\x0estreamVariable\x18\x01 \x01(\x0b\x32(.qm.grpc.qua.QuaProgram.VarRefExpressionH\x00\x12\x44\n\x0bstreamArray\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpressionH\x00\x12\x0b\n\x03loc\x18\x64 \x01(\tB\x0e\n\x0cstream_oneof\x1a\x62\n\x15\x41rrayLengthExpression\x12<\n\x05\x61rray\x18\x01 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1aV\n\x10VarRefExpression\x12\x0e\n\x04name\x18\x02 \x01(\tH\x00\x12\x12\n\x08ioNumber\x18\x03 \x01(\x05H\x00\x12\x0b\n\x03loc\x18\x64 \x01(\tB\x0b\n\tvar_oneofJ\x04\x08\x01\x10\x02\x1a\x39\n\x16StructVarRefExpression\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03loc\x18\x64 \x01(\tJ\x04\x08\x01\x10\x02\x1a\x43\n\x1b\x45xternalStreamRefExpression\x12\x11\n\tstream_id\x18\x02 \x01(\r\x12\x0b\n\x03loc\x18\x64 \x01(\tJ\x04\x08\x01\x10\x02\x1a\x8e\x01\n\x15\x41rrayVarRefExpression\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x46\n\tstructVar\x18\x03 \x01(\x0b\x32..qm.grpc.qua.QuaProgram.StructVarRefExpressionH\x00\x88\x01\x01\x12\x0b\n\x03loc\x18\x64 \x01(\tB\x0c\n\n_structVarJ\x04\x08\x01\x10\x02\x1a\xa8\x01\n\x16\x41rrayCellRefExpression\x12?\n\x08\x61rrayVar\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpression\x12:\n\x05index\x18\x03 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x0b\n\x03loc\x18\x64 \x01(\tJ\x04\x08\x01\x10\x02\x1a[\n\x11LiteralExpression\x12\r\n\x05value\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.qm.grpc.qua.QuaProgram.Type\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\xe9\x02\n\x10\x42inaryExpression\x12\x43\n\x02op\x18\x01 \x01(\x0e\x32\x37.qm.grpc.qua.QuaProgram.BinaryExpression.BinaryOperator\x12\x39\n\x04left\x18\x04 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12:\n\x05right\x18\x05 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x0b\n\x03loc\x18\x64 \x01(\t\"\x8b\x01\n\x0e\x42inaryOperator\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\x07\n\x03SUB\x10\x01\x12\x08\n\x04MULT\x10\x02\x12\x07\n\x03\x44IV\x10\x03\x12\x07\n\x03\x41ND\x10\x04\x12\x06\n\x02OR\x10\x05\x12\x07\n\x03XOR\x10\x06\x12\x06\n\x02LT\x10\x07\x12\x07\n\x03LET\x10\x08\x12\x06\n\x02GT\x10\t\x12\x07\n\x03GET\x10\n\x12\x06\n\x02\x45Q\x10\x0b\x12\x07\n\x03SHL\x10\x0c\x12\x07\n\x03SHR\x10\r\x1a\xb8\x02\n\x15LibFunctionExpression\x12\x14\n\x0c\x66unctionName\x18\x01 \x01(\t\x12I\n\targuments\x18\x02 \x03(\x0b\x32\x36.qm.grpc.qua.QuaProgram.LibFunctionExpression.Argument\x12\x13\n\x0blibraryName\x18\x03 \x01(\t\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\x9b\x01\n\x08\x41rgument\x12=\n\x06scalar\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpressionH\x00\x12>\n\x05\x61rray\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpressionH\x00\x42\x10\n\x0e\x61rgument_oneof\x1a\xd7\x05\n\x12\x46unctionExpression\x12\x45\n\x03\x61nd\x18\x01 \x01(\x0b\x32\x36.qm.grpc.qua.QuaProgram.FunctionExpression.AndFunctionH\x00\x12\x43\n\x02or\x18\x02 \x01(\x0b\x32\x35.qm.grpc.qua.QuaProgram.FunctionExpression.OrFunctionH\x00\x12\x45\n\x03xor\x18\x03 \x01(\x0b\x32\x36.qm.grpc.qua.QuaProgram.FunctionExpression.XorFunctionH\x00\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1a\xa9\x01\n\x16ScalarOrVectorArgument\x12=\n\x06scalar\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpressionH\x00\x12>\n\x05\x61rray\x18\x02 \x01(\x0b\x32-.qm.grpc.qua.QuaProgram.ArrayVarRefExpressionH\x00\x42\x10\n\x0e\x61rgument_oneof\x1a`\n\x0b\x41ndFunction\x12Q\n\x06values\x18\x01 \x03(\x0b\x32\x41.qm.grpc.qua.QuaProgram.FunctionExpression.ScalarOrVectorArgument\x1a_\n\nOrFunction\x12Q\n\x06values\x18\x01 \x03(\x0b\x32\x41.qm.grpc.qua.QuaProgram.FunctionExpression.ScalarOrVectorArgument\x1a`\n\x0bXorFunction\x12Q\n\x06values\x18\x01 \x03(\x0b\x32\x41.qm.grpc.qua.QuaProgram.FunctionExpression.ScalarOrVectorArgumentB\x10\n\x0e\x66unction_oneof\x1a^\n\x13\x42roadcastExpression\x12:\n\x05value\x18\x01 \x01(\x0b\x32+.qm.grpc.qua.QuaProgram.AnyScalarExpression\x12\x0b\n\x03loc\x18\x64 \x01(\t\x1ar\n\x16GlobalVarRefExpression\x12\x0b\n\x03loc\x18\x64 \x01(\t\x12\x0c\n\x04\x62its\x18\x01 \x03(\x05\x12=\n\toperation\x18\x02 \x01(\x0e\x32*.qm.grpc.qua.QuaProgram.GlobalVarOperation\x1aG\n\rDynamicConfig\x12\x0f\n\x07version\x18\x01 \x01(\r\x12%\n\x04root\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"#\n\x04Type\x12\x07\n\x03INT\x10\x00\x12\x08\n\x04\x42OOL\x10\x01\x12\x08\n\x04REAL\x10\x02\"\'\n\tDirection\x12\x0c\n\x08INCOMING\x10\x00\x12\x0c\n\x08OUTGOING\x10\x01\"7\n\x12GlobalVarOperation\x12\x08\n\x04read\x10\x00\x12\x0e\n\nread_shift\x10\x01\x12\x07\n\x03xor\x10\x02\"O\n\x11QuaResultAnalysis\x12\x0f\n\x07version\x18\x01 \x01(\r\x12)\n\x05model\x18\x03 \x03(\x0b\x32\x1a.google.protobuf.ListValueB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.pb.inc_qua_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['useExperimentalCalculationCompiler']._loaded_options = None
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['useExperimentalCalculationCompiler']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['optimizeMergeCodeExecution']._loaded_options = None
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['optimizeMergeCodeExecution']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['optimizeWriteReadCommands']._loaded_options = None
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['optimizeWriteReadCommands']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['skipOptimizations']._loaded_options = None
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['skipOptimizations']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['strict']._loaded_options = None
  _globals['_QUAPROGRAM_COMPILEROPTIONS'].fields_by_name['strict']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_ANYSTATEMENT'].fields_by_name['qrun']._loaded_options = None
  _globals['_QUAPROGRAM_ANYSTATEMENT'].fields_by_name['qrun']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_PLAYSTATEMENT'].fields_by_name['pulse']._loaded_options = None
  _globals['_QUAPROGRAM_PLAYSTATEMENT'].fields_by_name['pulse']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_MEASUREOUTPUT'].fields_by_name['variable']._loaded_options = None
  _globals['_QUAPROGRAM_MEASUREOUTPUT'].fields_by_name['variable']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_MEASURESTATEMENT'].fields_by_name['outputs']._loaded_options = None
  _globals['_QUAPROGRAM_MEASURESTATEMENT'].fields_by_name['outputs']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT'].fields_by_name['variable']._loaded_options = None
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT'].fields_by_name['variable']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT'].fields_by_name['save']._loaded_options = None
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT'].fields_by_name['save']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM_SAVESTATEMENT'].fields_by_name['variable']._loaded_options = None
  _globals['_QUAPROGRAM_SAVESTATEMENT'].fields_by_name['variable']._serialized_options = b'\030\001'
  _globals['_QUAPROGRAM']._serialized_start=184
  _globals['_QUAPROGRAM']._serialized_end=19180
  _globals['_QUAPROGRAM_COMPILEROPTIONS']._serialized_start=528
  _globals['_QUAPROGRAM_COMPILEROPTIONS']._serialized_end=738
  _globals['_QUAPROGRAM_SCRIPT']._serialized_start=741
  _globals['_QUAPROGRAM_SCRIPT']._serialized_end=944
  _globals['_QUAPROGRAM_VARDECLARATION']._serialized_start=947
  _globals['_QUAPROGRAM_VARDECLARATION']._serialized_end=1252
  _globals['_QUAPROGRAM_VARDECLARATION_STRUCTMEMBER']._serialized_start=1206
  _globals['_QUAPROGRAM_VARDECLARATION_STRUCTMEMBER']._serialized_end=1252
  _globals['_QUAPROGRAM_EXTERNALSTREAMDECLARATION']._serialized_start=1255
  _globals['_QUAPROGRAM_EXTERNALSTREAMDECLARATION']._serialized_end=1418
  _globals['_QUAPROGRAM_QUANTUMELEMENTREFERENCE']._serialized_start=1420
  _globals['_QUAPROGRAM_QUANTUMELEMENTREFERENCE']._serialized_end=1472
  _globals['_QUAPROGRAM_PULSEREFERENCE']._serialized_start=1474
  _globals['_QUAPROGRAM_PULSEREFERENCE']._serialized_end=1517
  _globals['_QUAPROGRAM_INTEGRATIONWEIGHTREFERENCE']._serialized_start=1519
  _globals['_QUAPROGRAM_INTEGRATIONWEIGHTREFERENCE']._serialized_end=1574
  _globals['_QUAPROGRAM_ANYSTATEMENT']._serialized_start=1577
  _globals['_QUAPROGRAM_ANYSTATEMENT']._serialized_end=3816
  _globals['_QUAPROGRAM_AMPMULTIPLIER']._serialized_start=3819
  _globals['_QUAPROGRAM_AMPMULTIPLIER']._serialized_end=4075
  _globals['_QUAPROGRAM_RAMPTOZEROSTATEMENT']._serialized_start=4078
  _globals['_QUAPROGRAM_RAMPTOZEROSTATEMENT']._serialized_end=4221
  _globals['_QUAPROGRAM_RAMPPULSE']._serialized_start=4223
  _globals['_QUAPROGRAM_RAMPPULSE']._serialized_end=4307
  _globals['_QUAPROGRAM_CHIRP']._serialized_start=4310
  _globals['_QUAPROGRAM_CHIRP']._serialized_end=4666
  _globals['_QUAPROGRAM_CHIRP_UNITS']._serialized_start=4557
  _globals['_QUAPROGRAM_CHIRP_UNITS']._serialized_end=4658
  _globals['_QUAPROGRAM_PLAYSTATEMENT']._serialized_start=4669
  _globals['_QUAPROGRAM_PLAYSTATEMENT']._serialized_end=5350
  _globals['_QUAPROGRAM_UPDATEFREQUENCYSTATEMENT']._serialized_start=5353
  _globals['_QUAPROGRAM_UPDATEFREQUENCYSTATEMENT']._serialized_end=5656
  _globals['_QUAPROGRAM_UPDATEFREQUENCYSTATEMENT_UNITS']._serialized_start=5605
  _globals['_QUAPROGRAM_UPDATEFREQUENCYSTATEMENT_UNITS']._serialized_end=5656
  _globals['_QUAPROGRAM_UPDATEOSCILLATORFREQUENCYSTATEMENT']._serialized_start=5659
  _globals['_QUAPROGRAM_UPDATEOSCILLATORFREQUENCYSTATEMENT']._serialized_end=6157
  _globals['_QUAPROGRAM_UPDATEOSCILLATORFREQUENCYSTATEMENT_UNITS']._serialized_start=6041
  _globals['_QUAPROGRAM_UPDATEOSCILLATORFREQUENCYSTATEMENT_UNITS']._serialized_end=6083
  _globals['_QUAPROGRAM_UPDATEOSCILLATORFREQUENCYSTATEMENT_UPDATECOMPONENTSELECTION']._serialized_start=6085
  _globals['_QUAPROGRAM_UPDATEOSCILLATORFREQUENCYSTATEMENT_UPDATECOMPONENTSELECTION']._serialized_end=6157
  _globals['_QUAPROGRAM_SETDCOFFSETSTATEMENT']._serialized_start=6160
  _globals['_QUAPROGRAM_SETDCOFFSETSTATEMENT']._serialized_end=6343
  _globals['_QUAPROGRAM_ZROTATIONSTATEMENT']._serialized_start=6346
  _globals['_QUAPROGRAM_ZROTATIONSTATEMENT']._serialized_end=6500
  _globals['_QUAPROGRAM_RESETFRAMESTATEMENT']._serialized_start=6502
  _globals['_QUAPROGRAM_RESETFRAMESTATEMENT']._serialized_end=6597
  _globals['_QUAPROGRAM_FASTFRAMEROTATIONSTATEMENT']._serialized_start=6600
  _globals['_QUAPROGRAM_FASTFRAMEROTATIONSTATEMENT']._serialized_end=6822
  _globals['_QUAPROGRAM_ALIGNSTATEMENT']._serialized_start=6824
  _globals['_QUAPROGRAM_ALIGNSTATEMENT']._serialized_end=6914
  _globals['_QUAPROGRAM_MEASUREPROCESS']._serialized_start=6917
  _globals['_QUAPROGRAM_MEASUREPROCESS']._serialized_end=7074
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS']._serialized_start=7077
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS']._serialized_end=8969
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_BAREINTEGRATION']._serialized_start=7673
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_BAREINTEGRATION']._serialized_end=7847
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_DUALBAREINTEGRATION']._serialized_start=7850
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_DUALBAREINTEGRATION']._serialized_end=8128
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_DEMODINTEGRATION']._serialized_start=8131
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_DEMODINTEGRATION']._serialized_end=8306
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_DUALDEMODINTEGRATION']._serialized_start=8309
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_DUALDEMODINTEGRATION']._serialized_end=8588
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_RAWTIMETAGGING']._serialized_start=8591
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_RAWTIMETAGGING']._serialized_end=8771
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_HIGHRESTIMETAGGING']._serialized_start=8774
  _globals['_QUAPROGRAM_ANALOGMEASUREPROCESS_HIGHRESTIMETAGGING']._serialized_end=8958
  _globals['_QUAPROGRAM_DIGITALMEASUREPROCESS']._serialized_start=8972
  _globals['_QUAPROGRAM_DIGITALMEASUREPROCESS']._serialized_end=9477
  _globals['_QUAPROGRAM_DIGITALMEASUREPROCESS_RAWTIMETAGGING']._serialized_start=9175
  _globals['_QUAPROGRAM_DIGITALMEASUREPROCESS_RAWTIMETAGGING']._serialized_end=9355
  _globals['_QUAPROGRAM_DIGITALMEASUREPROCESS_COUNTING']._serialized_start=9357
  _globals['_QUAPROGRAM_DIGITALMEASUREPROCESS_COUNTING']._serialized_end=9466
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET']._serialized_start=9480
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET']._serialized_end=10304
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET_SCALARPROCESSTARGET']._serialized_start=9697
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET_SCALARPROCESSTARGET']._serialized_end=9859
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET_VECTORPROCESSTARGET']._serialized_start=9862
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET_VECTORPROCESSTARGET']._serialized_end=10025
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET_TIMEDIVISION']._serialized_start=10028
  _globals['_QUAPROGRAM_ANALOGPROCESSTARGET_TIMEDIVISION']._serialized_end=10287
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION']._serialized_start=10307
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION']._serialized_end=10507
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION_SLICED']._serialized_start=10329
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION_SLICED']._serialized_end=10375
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION_ACCUMULATED']._serialized_start=10377
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION_ACCUMULATED']._serialized_end=10428
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION_MOVINGWINDOW']._serialized_start=10430
  _globals['_QUAPROGRAM_ANALOGTIMEDIVISION_MOVINGWINDOW']._serialized_end=10507
  _globals['_QUAPROGRAM_MEASUREOUTPUT']._serialized_start=10510
  _globals['_QUAPROGRAM_MEASUREOUTPUT']._serialized_end=10905
  _globals['_QUAPROGRAM_MEASUREOUTPUT_TARGET']._serialized_start=10756
  _globals['_QUAPROGRAM_MEASUREOUTPUT_TARGET']._serialized_end=10905
  _globals['_QUAPROGRAM_MEASURESTATEMENT']._serialized_start=10908
  _globals['_QUAPROGRAM_MEASURESTATEMENT']._serialized_end=11275
  _globals['_QUAPROGRAM_WAITSTATEMENT']._serialized_start=11278
  _globals['_QUAPROGRAM_WAITSTATEMENT']._serialized_end=11426
  _globals['_QUAPROGRAM_ARBITRARYSTATEMENT']._serialized_start=11428
  _globals['_QUAPROGRAM_ARBITRARYSTATEMENT']._serialized_end=11511
  _globals['_QUAPROGRAM_ARBITRARYCONTEXTSTATEMENT']._serialized_start=11514
  _globals['_QUAPROGRAM_ARBITRARYCONTEXTSTATEMENT']._serialized_end=11678
  _globals['_QUAPROGRAM_WAITFORTRIGGERSTATEMENT']._serialized_start=11681
  _globals['_QUAPROGRAM_WAITFORTRIGGERSTATEMENT']._serialized_end=12103
  _globals['_QUAPROGRAM_WAITFORTRIGGERSTATEMENT_ELEMENTOUTPUT']._serialized_start=12045
  _globals['_QUAPROGRAM_WAITFORTRIGGERSTATEMENT_ELEMENTOUTPUT']._serialized_end=12093
  _globals['_QUAPROGRAM_UPDATECORRECTIONSTATEMENT']._serialized_start=12106
  _globals['_QUAPROGRAM_UPDATECORRECTIONSTATEMENT']._serialized_end=12263
  _globals['_QUAPROGRAM_RESETPHASESTATEMENT']._serialized_start=12265
  _globals['_QUAPROGRAM_RESETPHASESTATEMENT']._serialized_end=12360
  _globals['_QUAPROGRAM_RESETGLOBALPHASESTATEMENT']._serialized_start=12362
  _globals['_QUAPROGRAM_RESETGLOBALPHASESTATEMENT']._serialized_end=12402
  _globals['_QUAPROGRAM_SENDTOEXTERNALSTREAMSTATEMENT']._serialized_start=12405
  _globals['_QUAPROGRAM_SENDTOEXTERNALSTREAMSTATEMENT']._serialized_end=12582
  _globals['_QUAPROGRAM_RECEIVEFROMEXTERNALSTREAMSTATEMENT']._serialized_start=12585
  _globals['_QUAPROGRAM_RECEIVEFROMEXTERNALSTREAMSTATEMENT']._serialized_end=12767
  _globals['_QUAPROGRAM_GLOBALVARIABLEASSIGNMENTSTATEMENT']._serialized_start=12769
  _globals['_QUAPROGRAM_GLOBALVARIABLEASSIGNMENTSTATEMENT']._serialized_end=12878
  _globals['_QUAPROGRAM_CORRECTION']._serialized_start=12881
  _globals['_QUAPROGRAM_CORRECTION']._serialized_end=13134
  _globals['_QUAPROGRAM_SYNCSTATEMENT']._serialized_start=13136
  _globals['_QUAPROGRAM_SYNCSTATEMENT']._serialized_end=13164
  _globals['_QUAPROGRAM_ELSEIF']._serialized_start=13167
  _globals['_QUAPROGRAM_ELSEIF']._serialized_end=13312
  _globals['_QUAPROGRAM_IFSTATEMENT']._serialized_start=13315
  _globals['_QUAPROGRAM_IFSTATEMENT']._serialized_end=13590
  _globals['_QUAPROGRAM_FORSTATEMENT']._serialized_start=13593
  _globals['_QUAPROGRAM_FORSTATEMENT']._serialized_end=13866
  _globals['_QUAPROGRAM_QRUNSTATEMENT']._serialized_start=13868
  _globals['_QUAPROGRAM_QRUNSTATEMENT']._serialized_end=13956
  _globals['_QUAPROGRAM_STRICTTIMINGSTATEMENT']._serialized_start=13958
  _globals['_QUAPROGRAM_STRICTTIMINGSTATEMENT']._serialized_end=14054
  _globals['_QUAPROGRAM_FOREACHSTATEMENT']._serialized_start=14057
  _globals['_QUAPROGRAM_FOREACHSTATEMENT']._serialized_end=14372
  _globals['_QUAPROGRAM_FOREACHSTATEMENT_VARIABLEWITHVALUES']._serialized_start=14230
  _globals['_QUAPROGRAM_FOREACHSTATEMENT_VARIABLEWITHVALUES']._serialized_end=14372
  _globals['_QUAPROGRAM_RUSSTATEMENT']._serialized_start=14375
  _globals['_QUAPROGRAM_RUSSTATEMENT']._serialized_end=14590
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT']._serialized_start=14593
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT']._serialized_end=14994
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT_TARGET']._serialized_start=10756
  _globals['_QUAPROGRAM_ASSIGNMENTSTATEMENT_TARGET']._serialized_end=10905
  _globals['_QUAPROGRAM_PAUSESTATEMENT']._serialized_start=14996
  _globals['_QUAPROGRAM_PAUSESTATEMENT']._serialized_end=15087
  _globals['_QUAPROGRAM_SAVESTATEMENT']._serialized_start=15090
  _globals['_QUAPROGRAM_SAVESTATEMENT']._serialized_end=15471
  _globals['_QUAPROGRAM_SAVESTATEMENT_SOURCE']._serialized_start=15260
  _globals['_QUAPROGRAM_SAVESTATEMENT_SOURCE']._serialized_end=15471
  _globals['_QUAPROGRAM_LOADWAVEFORMSTATEMENT']._serialized_start=15474
  _globals['_QUAPROGRAM_LOADWAVEFORMSTATEMENT']._serialized_end=15695
  _globals['_QUAPROGRAM_STATEMENTSCOLLECTION']._serialized_start=15697
  _globals['_QUAPROGRAM_STATEMENTSCOLLECTION']._serialized_end=15777
  _globals['_QUAPROGRAM_ANYSCALAREXPRESSION']._serialized_start=15780
  _globals['_QUAPROGRAM_ANYSCALAREXPRESSION']._serialized_end=16427
  _globals['_QUAPROGRAM_ADVANCEINPUTSTREAMSTATEMENT']._serialized_start=16430
  _globals['_QUAPROGRAM_ADVANCEINPUTSTREAMSTATEMENT']._serialized_end=16626
  _globals['_QUAPROGRAM_ARRAYLENGTHEXPRESSION']._serialized_start=16628
  _globals['_QUAPROGRAM_ARRAYLENGTHEXPRESSION']._serialized_end=16726
  _globals['_QUAPROGRAM_VARREFEXPRESSION']._serialized_start=16728
  _globals['_QUAPROGRAM_VARREFEXPRESSION']._serialized_end=16814
  _globals['_QUAPROGRAM_STRUCTVARREFEXPRESSION']._serialized_start=16816
  _globals['_QUAPROGRAM_STRUCTVARREFEXPRESSION']._serialized_end=16873
  _globals['_QUAPROGRAM_EXTERNALSTREAMREFEXPRESSION']._serialized_start=16875
  _globals['_QUAPROGRAM_EXTERNALSTREAMREFEXPRESSION']._serialized_end=16942
  _globals['_QUAPROGRAM_ARRAYVARREFEXPRESSION']._serialized_start=16945
  _globals['_QUAPROGRAM_ARRAYVARREFEXPRESSION']._serialized_end=17087
  _globals['_QUAPROGRAM_ARRAYCELLREFEXPRESSION']._serialized_start=17090
  _globals['_QUAPROGRAM_ARRAYCELLREFEXPRESSION']._serialized_end=17258
  _globals['_QUAPROGRAM_LITERALEXPRESSION']._serialized_start=17260
  _globals['_QUAPROGRAM_LITERALEXPRESSION']._serialized_end=17351
  _globals['_QUAPROGRAM_BINARYEXPRESSION']._serialized_start=17354
  _globals['_QUAPROGRAM_BINARYEXPRESSION']._serialized_end=17715
  _globals['_QUAPROGRAM_BINARYEXPRESSION_BINARYOPERATOR']._serialized_start=17576
  _globals['_QUAPROGRAM_BINARYEXPRESSION_BINARYOPERATOR']._serialized_end=17715
  _globals['_QUAPROGRAM_LIBFUNCTIONEXPRESSION']._serialized_start=17718
  _globals['_QUAPROGRAM_LIBFUNCTIONEXPRESSION']._serialized_end=18030
  _globals['_QUAPROGRAM_LIBFUNCTIONEXPRESSION_ARGUMENT']._serialized_start=17875
  _globals['_QUAPROGRAM_LIBFUNCTIONEXPRESSION_ARGUMENT']._serialized_end=18030
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION']._serialized_start=18033
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION']._serialized_end=18760
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_SCALARORVECTORARGUMENT']._serialized_start=18280
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_SCALARORVECTORARGUMENT']._serialized_end=18449
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_ANDFUNCTION']._serialized_start=18451
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_ANDFUNCTION']._serialized_end=18547
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_ORFUNCTION']._serialized_start=18549
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_ORFUNCTION']._serialized_end=18644
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_XORFUNCTION']._serialized_start=18646
  _globals['_QUAPROGRAM_FUNCTIONEXPRESSION_XORFUNCTION']._serialized_end=18742
  _globals['_QUAPROGRAM_BROADCASTEXPRESSION']._serialized_start=18762
  _globals['_QUAPROGRAM_BROADCASTEXPRESSION']._serialized_end=18856
  _globals['_QUAPROGRAM_GLOBALVARREFEXPRESSION']._serialized_start=18858
  _globals['_QUAPROGRAM_GLOBALVARREFEXPRESSION']._serialized_end=18972
  _globals['_QUAPROGRAM_DYNAMICCONFIG']._serialized_start=18974
  _globals['_QUAPROGRAM_DYNAMICCONFIG']._serialized_end=19045
  _globals['_QUAPROGRAM_TYPE']._serialized_start=19047
  _globals['_QUAPROGRAM_TYPE']._serialized_end=19082
  _globals['_QUAPROGRAM_DIRECTION']._serialized_start=19084
  _globals['_QUAPROGRAM_DIRECTION']._serialized_end=19123
  _globals['_QUAPROGRAM_GLOBALVAROPERATION']._serialized_start=19125
  _globals['_QUAPROGRAM_GLOBALVAROPERATION']._serialized_end=19180
  _globals['_QUARESULTANALYSIS']._serialized_start=19182
  _globals['_QUARESULTANALYSIS']._serialized_end=19261
# @@protoc_insertion_point(module_scope)



# This file provides type hints for the Qua DSL, allowing users to import them from `qm.qua.type_hints` instead of
# protected files.
from qm.qua._dsl.play import ChirpUnits
from qm.qua import ChirpType, StreamType
from qm.type_hinting.general import NumberT
from qm.qua._dsl.stream_processing.stream_processing import ResultStreamSource
from qm.qua._expressions import (
    NSize,
    QuaIO,
    Scalar,
    Vector,
    QuaScalar,
    QuaVariable,
    QuaArrayCell,
    QuaBroadcast,
    ScalarOfAnyType,
    VectorOfAnyType,
    QuaArrayInputStream,
    QuaScalarExpression,
    QuaLibFunctionOutput,
    QuaVariableInputStream,
)

__all__ = [
    "NumberT",
    "ChirpType",
    "ChirpUnits",
    "Scalar",
    "Vector",
    "QuaScalar",
    "ScalarOfAnyType",
    "VectorOfAnyType",
    "QuaVariable",
    "QuaArrayCell",
    "QuaLibFunctionOutput",
    "QuaBroadcast",
    "QuaScalarExpression",
    "QuaArrayInputStream",
    "QuaVariableInputStream",
    "QuaIO",
    "StreamType",
    "NSize",
    "ResultStreamSource",
]



import json
import datetime
from enum import Enum
from dataclasses import dataclass
from typing import Any, List, Tuple, cast

DtypeType = List[List[str]]


class InsertDirection(Enum):
    start = 1
    end = 2


@dataclass(frozen=True)
class PendingJobData:
    job_id: str
    position_in_queue: int
    time_added: datetime.datetime
    added_by: str


@dataclass
class JobNamedResult:
    data: bytes
    count_of_items: int
    output_name: str


@dataclass
class JobResultItemSchema:
    name: str
    bare_dtype: str
    shape: Tuple[int, ...]
    is_single: bool
    expected_count: int

    @property
    def dtype(self) -> DtypeType:
        return _parse_dtype(self.bare_dtype)


@dataclass
class JobStreamingState:
    job_id: str
    done: bool
    closed: bool
    has_dataloss: bool


@dataclass
class JobNamedResultHeader:
    count_so_far: int
    bare_dtype: str
    shape: tuple[int, ...]
    has_dataloss: bool
    has_execution_errors: bool

    @property
    def d_type(self) -> DtypeType:
        return _parse_dtype(self.bare_dtype)


def _parse_dtype(simple_dtype: str) -> DtypeType:
    def hinted_tuple_hook(obj: Any) -> Any:
        if "__tuple__" in obj:
            return tuple(obj["items"])
        else:
            return obj

    dtype = json.loads(simple_dtype, object_hook=hinted_tuple_hook)
    return cast(DtypeType, dtype)



from google.protobuf import wrappers_pb2 as _wrappers_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class JobExecutionStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNSET: _ClassVar[JobExecutionStatus]
    UNKNOWN: _ClassVar[JobExecutionStatus]
    PENDING: _ClassVar[JobExecutionStatus]
    RUNNING: _ClassVar[JobExecutionStatus]
    COMPLETED: _ClassVar[JobExecutionStatus]
    CANCELED: _ClassVar[JobExecutionStatus]
    LOADING: _ClassVar[JobExecutionStatus]
    ERROR: _ClassVar[JobExecutionStatus]
    PROCESSING: _ClassVar[JobExecutionStatus]
UNSET: JobExecutionStatus
UNKNOWN: JobExecutionStatus
PENDING: JobExecutionStatus
RUNNING: JobExecutionStatus
COMPLETED: JobExecutionStatus
CANCELED: JobExecutionStatus
LOADING: JobExecutionStatus
ERROR: JobExecutionStatus
PROCESSING: JobExecutionStatus

class Range(_message.Message):
    __slots__ = ("to",)
    FROM_FIELD_NUMBER: _ClassVar[int]
    TO_FIELD_NUMBER: _ClassVar[int]
    to: _wrappers_pb2.Int64Value
    def __init__(self, to: _Optional[_Union[_wrappers_pb2.Int64Value, _Mapping]] = ..., **kwargs) -> None: ...



from typing import Dict, Union, Literal, Protocol, cast, overload

from qm.exceptions import InvalidConfigError
from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.utils.protobuf_utils import which_one_of

FemTypes = Union[
    inc_qua_config_pb2.QuaConfig.ControllerDec,
    inc_qua_config_pb2.QuaConfig.OctoDacFemDec,
    inc_qua_config_pb2.QuaConfig.MicrowaveFemDec,
]


class ControllerConfigProtocol(Protocol):
    control_devices: Dict[str, inc_qua_config_pb2.QuaConfig.DeviceDec]
    mixers: Dict[str, inc_qua_config_pb2.QuaConfig.MixerDec]
    octaves: Dict[str, inc_qua_config_pb2.QuaConfig.Octave.Config]


class LogicalConfigProtocol(Protocol):
    elements: Dict[str, inc_qua_config_pb2.QuaConfig.ElementDec]
    oscillators: Dict[str, inc_qua_config_pb2.QuaConfig.Oscillator]
    pulses: Dict[str, inc_qua_config_pb2.QuaConfig.PulseDec]
    waveforms: Dict[str, inc_qua_config_pb2.QuaConfig.WaveformDec]
    digital_waveforms: Dict[str, inc_qua_config_pb2.QuaConfig.DigitalWaveformDec]
    integration_weights: Dict[str, inc_qua_config_pb2.QuaConfig.IntegrationWeightDec]


# TODO: See if the functions in this file can be moved to the config models when they are created.


@overload
def _get_correct_config_part(
    part: Literal["controller"], pb_config: inc_qua_config_pb2.QuaConfig
) -> Union[inc_qua_config_pb2.QuaConfig.ControllerConfig, inc_qua_config_pb2.QuaConfig.QuaConfigV1]:
    pass


@overload
def _get_correct_config_part(
    part: Literal["logical"], pb_config: inc_qua_config_pb2.QuaConfig
) -> Union[inc_qua_config_pb2.QuaConfig.LogicalConfig, inc_qua_config_pb2.QuaConfig.QuaConfigV1]:
    pass


def _get_correct_config_part(
    part: Literal["controller", "logical"], pb_config: inc_qua_config_pb2.QuaConfig
) -> Union[
    inc_qua_config_pb2.QuaConfig.ControllerConfig,
    inc_qua_config_pb2.QuaConfig.LogicalConfig,
    inc_qua_config_pb2.QuaConfig.QuaConfigV1,
]:
    _, config_version_inst = which_one_of(pb_config, "config_version")
    if isinstance(config_version_inst, inc_qua_config_pb2.QuaConfig.QuaConfigV2):
        if part == "controller":
            return config_version_inst.controller_config
        elif part == "logical":
            return config_version_inst.logical_config
    elif isinstance(config_version_inst, inc_qua_config_pb2.QuaConfig.QuaConfigV1):
        return config_version_inst
    else:
        raise ValueError("Received unknown config type: " + str(type(config_version_inst)))


def _unset_correct_config_part(part: Literal["controller", "logical"], pb_config: inc_qua_config_pb2.QuaConfig) -> None:
    _, config_version_inst = which_one_of(pb_config, "config_version")
    if isinstance(config_version_inst, inc_qua_config_pb2.QuaConfig.QuaConfigV2):
        if part == "controller":
            config_version_inst.ClearField("controller_config")
        elif part == "logical":
            config_version_inst.ClearField("logical_config")
    elif isinstance(config_version_inst, inc_qua_config_pb2.QuaConfig.QuaConfigV1):
        pass
    else:
        raise ValueError("Received unknown config type: " + str(type(config_version_inst)))


def get_controller_pb_config(pb_config: inc_qua_config_pb2.QuaConfig) -> inc_qua_config_pb2.QuaConfig.ControllerConfig:
    return cast(inc_qua_config_pb2.QuaConfig.ControllerConfig, _get_correct_config_part("controller", pb_config))


def get_logical_pb_config(pb_config: inc_qua_config_pb2.QuaConfig) -> inc_qua_config_pb2.QuaConfig.LogicalConfig:
    return cast(inc_qua_config_pb2.QuaConfig.LogicalConfig, _get_correct_config_part("logical", pb_config))


def unset_controller_pb_config(pb_config: inc_qua_config_pb2.QuaConfig) -> None:
    _unset_correct_config_part("controller", pb_config)


def unset_logical_pb_config(pb_config: inc_qua_config_pb2.QuaConfig) -> None:
    _unset_correct_config_part("logical", pb_config)


def get_fem_config_instance(fem_ref: inc_qua_config_pb2.QuaConfig.FEMTypes) -> FemTypes:
    _, config = which_one_of(fem_ref, "fem_type_one_of")
    if not isinstance(
        config,
        (
            inc_qua_config_pb2.QuaConfig.ControllerDec,
            inc_qua_config_pb2.QuaConfig.OctoDacFemDec,
            inc_qua_config_pb2.QuaConfig.MicrowaveFemDec,
        ),
    ):
        raise InvalidConfigError(f"FEM type {type(config)} is not supported")
    return config


def get_fem_config(
    pb_config: inc_qua_config_pb2.QuaConfig,
    port: Union[
        inc_qua_config_pb2.QuaConfig.DacPortReference,
        inc_qua_config_pb2.QuaConfig.AdcPortReference,
        inc_qua_config_pb2.QuaConfig.PortReference,
    ],
) -> FemTypes:
    controller_config = get_controller_pb_config(pb_config)
    if port.controller not in controller_config.controlDevices:
        raise InvalidConfigError("Controller not found")
    controller = controller_config.controlDevices[port.controller]
    if port.fem not in controller.fems:
        raise InvalidConfigError("FEM not found")

    fem_ref = controller.fems[port.fem]
    config = get_fem_config_instance(fem_ref)
    return config


def element_has_mix_inputs(element: inc_qua_config_pb2.QuaConfig.ElementDec) -> bool:
    _, inputs_inst = which_one_of(element, "element_inputs_one_of")
    return isinstance(inputs_inst, inc_qua_config_pb2.QuaConfig.MixInputs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/pb/qm_manager.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/pb/qm_manager.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from qm.grpc.qm.pb import inc_qua_config_pb2 as qm_dot_pb_dot_inc__qua__config__pb2
from qm.grpc.qm.pb import general_messages_pb2 as qm_dot_pb_dot_general__messages__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16qm/pb/qm_manager.proto\x12\x12qm.grpc.qm_manager\x1a\x1aqm/pb/inc_qua_config.proto\x1a\x1cqm/pb/general_messages.proto\"\xc2\x01\n\x19OpenQuantumMachineRequest\x12-\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x12\x0f\n\x05never\x18\x02 \x01(\x08H\x00\x12\x10\n\x06\x61lways\x18\x03 \x01(\x08H\x00\x12\x12\n\x08ifNeeded\x18\x04 \x01(\x08H\x00\x12$\n\x1ckeep_dc_offsets_when_closing\x18\x05 \x01(\x08\x42\x19\n\x17oneOfCloseOtherMachines\"\x99\x02\n\x1aOpenQuantumMachineResponse\x12\x11\n\tmachineID\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12K\n\x16\x63onfigValidationErrors\x18\x03 \x03(\x0b\x32+.qm.grpc.qm_manager.ConfigValidationMessage\x12O\n\x18physicalValidationErrors\x18\x04 \x03(\x0b\x32-.qm.grpc.qm_manager.PhysicalValidationMessage\x12\x39\n\x0eopenQmWarnings\x18\x05 \x03(\x0b\x32!.qm.grpc.qm_manager.OpenQmWarning\"/\n\x1a\x43loseQuantumMachineRequest\x12\x11\n\tmachineID\x18\x01 \x01(\t\"f\n\x1b\x43loseQuantumMachineResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x02 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\"-\n\x18GetQuantumMachineRequest\x12\x11\n\tmachineID\x18\x01 \x01(\t\"\xa6\x01\n\x19GetQuantumMachineResponse\x12\x11\n\tmachineID\x18\x01 \x01(\t\x12-\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x1d.qm.grpc.qua_config.QuaConfig\x12\x0f\n\x07success\x18\x03 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x04 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\")\n\x14GetRunningJobRequest\x12\x11\n\tmachineID\x18\x01 \x01(\t\"9\n\x15GetRunningJobResponse\x12\x11\n\tmachineID\x18\x01 \x01(\t\x12\r\n\x05jobId\x18\x02 \x01(\t\"5\n\x1fListOpenQuantumMachinesResponse\x12\x12\n\nmachineIDs\x18\x01 \x03(\t\"j\n\x1f\x43loseAllQuantumMachinesResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x36\n\x06\x65rrors\x18\x02 \x03(\x0b\x32&.qm.grpc.general_messages.ErrorMessage\"M\n\x16GetControllersResponse\x12\x33\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x1e.qm.grpc.qm_manager.Controller\"D\n\nController\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x0btemperature\x18\x02 \x01(\x01H\x00\x88\x01\x01\x42\x0e\n\x0c_temperature\"~\n\x17\x43onfigValidationMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x35\n\x05level\x18\x04 \x01(\x0e\x32&.qm.grpc.general_messages.MessageLevel\"\x80\x01\n\x19PhysicalValidationMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x35\n\x05level\x18\x04 \x01(\x0e\x32&.qm.grpc.general_messages.MessageLevel\".\n\rOpenQmWarning\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\tB+\n\x11qm.grpc.qmManagerB\x0eQmManagerProtoP\x01\xa2\x02\x03HLWb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.pb.qm_manager_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021qm.grpc.qmManagerB\016QmManagerProtoP\001\242\002\003HLW'
  _globals['_OPENQUANTUMMACHINEREQUEST']._serialized_start=105
  _globals['_OPENQUANTUMMACHINEREQUEST']._serialized_end=299
  _globals['_OPENQUANTUMMACHINERESPONSE']._serialized_start=302
  _globals['_OPENQUANTUMMACHINERESPONSE']._serialized_end=583
  _globals['_CLOSEQUANTUMMACHINEREQUEST']._serialized_start=585
  _globals['_CLOSEQUANTUMMACHINEREQUEST']._serialized_end=632
  _globals['_CLOSEQUANTUMMACHINERESPONSE']._serialized_start=634
  _globals['_CLOSEQUANTUMMACHINERESPONSE']._serialized_end=736
  _globals['_GETQUANTUMMACHINEREQUEST']._serialized_start=738
  _globals['_GETQUANTUMMACHINEREQUEST']._serialized_end=783
  _globals['_GETQUANTUMMACHINERESPONSE']._serialized_start=786
  _globals['_GETQUANTUMMACHINERESPONSE']._serialized_end=952
  _globals['_GETRUNNINGJOBREQUEST']._serialized_start=954
  _globals['_GETRUNNINGJOBREQUEST']._serialized_end=995
  _globals['_GETRUNNINGJOBRESPONSE']._serialized_start=997
  _globals['_GETRUNNINGJOBRESPONSE']._serialized_end=1054
  _globals['_LISTOPENQUANTUMMACHINESRESPONSE']._serialized_start=1056
  _globals['_LISTOPENQUANTUMMACHINESRESPONSE']._serialized_end=1109
  _globals['_CLOSEALLQUANTUMMACHINESRESPONSE']._serialized_start=1111
  _globals['_CLOSEALLQUANTUMMACHINESRESPONSE']._serialized_end=1217
  _globals['_GETCONTROLLERSRESPONSE']._serialized_start=1219
  _globals['_GETCONTROLLERSRESPONSE']._serialized_end=1296
  _globals['_CONTROLLER']._serialized_start=1298
  _globals['_CONTROLLER']._serialized_end=1366
  _globals['_CONFIGVALIDATIONMESSAGE']._serialized_start=1368
  _globals['_CONFIGVALIDATIONMESSAGE']._serialized_end=1494
  _globals['_PHYSICALVALIDATIONMESSAGE']._serialized_start=1497
  _globals['_PHYSICALVALIDATIONMESSAGE']._serialized_end=1625
  _globals['_OPENQMWARNING']._serialized_start=1627
  _globals['_OPENQMWARNING']._serialized_end=1673
# @@protoc_insertion_point(module_scope)



from qm.qua.extensions.qua_iterators.qua_zip import QuaZip
from qm.qua.extensions.qua_iterators.qua_product import QuaProduct
from qm.qua.extensions.qua_iterators.qua_iterators import QuaIterable, QuaIterableRange
from qm.qua.extensions.qua_iterators.qua_native_iterators import NativeIterable, NativeIterableRange

__all__ = [
    "QuaIterable",
    "QuaIterableRange",
    "NativeIterable",
    "NativeIterableRange",
    "QuaZip",
    "QuaProduct",
]



from qm.grpc.qm.grpc.v2 import job_api_pb2
from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.job_api.generic_apis import ElementGenericApi


class ElementPortApi(ElementGenericApi):
    def __init__(self, connection_details: ConnectionDetails, job_id: str, element_id: str, port_name: str) -> None:
        super().__init__(connection_details, job_id, element_id)
        self._port = port_name


class AnalogOutputApi(ElementPortApi):
    def set_dc_offset(self, value: float) -> None:
        request = job_api_pb2.SetInputDcOffsetRequest(
            job_id=self._id, qe=self._element_id, port=self._port, offset=value
        )
        self._run(self._stub.SetInputDcOffset, request, timeout=self._timeout)

    def get_dc_offset(self) -> float:
        request = job_api_pb2.GetInputDcOffsetRequest(job_id=self._id, qe=self._element_id, port=self._port)
        response: job_api_pb2.GetInputDcOffsetResponse.GetInputDcOffsetResponseSuccess = self._run(
            self._stub.GetInputDcOffset, request, timeout=self._timeout
        )
        return response.offset


class MwOutputApi(ElementGenericApi):
    def set_oscillator_frequency(self, frequency: float) -> None:
        request = job_api_pb2.SetOscillatorFrequencyRequest(
            job_id=self._id,
            qe=self._element_id,
            new_frequency_hz=frequency,
            update_component=job_api_pb2.SetOscillatorFrequencyRequest.UpdateComponentSelection.downconverter,
        )
        self._run(self._stub.SetOscillatorFrequency, request, timeout=self._timeout)


class DigitalInputApi(ElementPortApi):
    def set_delay(self, value: int) -> None:
        request = job_api_pb2.SetDigitalDelayRequest(job_id=self._id, qe=self._element_id, port=self._port, delay=value)
        self._run(self._stub.SetDigitalDelay, request, timeout=self._timeout)

    def get_delay(self) -> int:
        request = job_api_pb2.GetDigitalDelayRequest(job_id=self._id, qe=self._element_id, port=self._port)
        response: job_api_pb2.GetDigitalDelayResponse.GetDigitalDelayResponseSuccess = self._run(
            self._stub.GetDigitalDelay, request, timeout=self._timeout
        )
        return response.delay

    def set_buffer(self, value: int) -> None:
        request = job_api_pb2.SetDigitalBufferRequest(
            job_id=self._id, qe=self._element_id, port=self._port, buffer=value
        )
        self._run(self._stub.SetDigitalBuffer, request, timeout=self._timeout)

    def get_buffer(self) -> int:
        request = job_api_pb2.GetDigitalBufferRequest(job_id=self._id, qe=self._element_id, port=self._port)
        response: job_api_pb2.GetDigitalBufferResponse.GetDigitalBufferResponseSuccess = self._run(
            self._stub.GetDigitalBuffer, request, timeout=self._timeout
        )
        return response.buffer



import logging
import threading
from typing import Callable, Optional

import grpc

from qm.grpc.octave.v1 import api_pb2
from qm.octave_sdk._octave_client import OctaveClient, ExploreResult, MonitorResult

logger = logging.getLogger("qm")


NUM_CONNECTION_RETRIES = 3


class HealthClient:
    # Creates a constant gRPC stream to the Server, will call the "health_update_callback" on new health result
    # health_update_callback args should be (ExploreResult, MonitorResult)
    def __init__(
        self,
        interval: int,
        client: OctaveClient,
        health_update_callback: Callable[[ExploreResult, MonitorResult], None],
    ):
        self._interval = interval
        self._client = client
        self._callback = health_update_callback
        self._stream_running = False
        self._grace_stop = False
        self._connection_lost = 0

        self._thread: Optional[threading.Thread] = None
        self._stop_event = threading.Event()

    def start(self) -> None:
        self._stop_event.clear()
        self._grace_stop = False
        self._thread = threading.Thread(target=self._health_loop, daemon=True)
        self._thread.start()

    def stop(self) -> None:
        self._grace_stop = True
        self._stop_event.set()
        if self._thread and self._thread.is_alive():
            self._thread.join(timeout=1)

    def _health_loop(self) -> None:
        while not self._stop_event.is_set():
            try:
                self._health_monitor()
            except (TimeoutError, ConnectionRefusedError):
                pass
            except grpc.RpcError as e:
                # Get status code from gRPC exception
                status_code = e.code() if hasattr(e, "code") else None

                if status_code == grpc.StatusCode.UNIMPLEMENTED:
                    # In case of health monitor not implemented, we do not want to print messages of disconnection
                    logger.warning(f'Octave "{self._client.octave_name}" does not support live monitoring')
                    self._stream_running = False
                    return
                # all other gRPC errors are ignored (mirrors old async logic)
            finally:
                if not self._grace_stop and self._stream_running:
                    self._connection_lost += 1
                    logger.error(f'Octave "{self._client.octave_name}" lost monitor connection')
                    if self._connection_lost == NUM_CONNECTION_RETRIES:
                        logger.error(f"Failed {NUM_CONNECTION_RETRIES} times to connect Octave aborting")

    def _health_monitor(self) -> None:
        self._stream_running = False
        try:
            self._client._service.GetVersion(api_pb2.GetVersionRequest())
            self._health_answers = self._client._service.Health(
                iter([api_pb2.HealthRequest(monitor_interval_seconds=self._interval, stop_stream=False)])
            )
            self._stream_running = True
            if self._connection_lost:
                self._connection_lost = 0
                logger.info(f'Octave "{self._client.octave_name}" restored monitor connection')

            # FYI infinite loop here until server stream ends or task cancel
            for response in self._health_answers:
                self._callback(ExploreResult(response.explore), MonitorResult(response.monitor))

        except (TimeoutError, ConnectionRefusedError):
            pass
        except grpc.RpcError as e:
            # Get status code from gRPC exception
            status_code = e.code() if hasattr(e, "code") else None

            # cancelled exception is ignored
            if status_code != grpc.StatusCode.CANCELLED:
                raise

    def run_once(self) -> None:
        explore_result = self._client.explore()
        monitor_result = self._client.monitor()
        self._callback(explore_result, monitor_result)



import uuid
import warnings
from typing import Union, Optional, MutableMapping

from qm.utils import deprecation_message
from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.utils.protobuf_utils import which_one_of
from qm.api.models.capabilities import OPX_FEM_IDX, QopCaps, ServerCapabilities
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.program._dict_to_pb_converter.converters.pulse_converter import PulseConverter
from qm.program._dict_to_pb_converter.converters.octave_converter import OctaveConverter
from qm.program._dict_to_pb_converter.converters.waveform_converter import WaveformConverter
from qm.program._dict_to_pb_converter.converters.oscillator_converter import OscillatorConverter
from qm.utils.config_utils import get_logical_pb_config, get_fem_config_instance, get_controller_pb_config
from qm.program._dict_to_pb_converter.converters.mixer_correction_converter import MixerCorrectionConverter
from qm.program._dict_to_pb_converter.converters.integration_weights_converter import IntegrationWeightsConverter
from qm.program._dict_to_pb_converter.converters.element_converter import ElementConverter, StickyElementIsNotSupported
from qm.type_hinting.config_types import FullQuaConfig, LogicalQuaConfig, ControllerQuaConfig, DigitalWaveformConfigType
from qm.program._dict_to_pb_converter.converters.control_device_converter.control_device_converter import (
    ControlDeviceConverter,
)
from qm.exceptions import (
    NoMatchingOctavePort,
    ConfigValidationException,
    OctaveUnsupportedOnUpdate,
    ConfigurationLockedByOctave,
    ElementInputConnectionAmbiguity,
    ElementOutputConnectionAmbiguity,
)


class DictToQuaConfigConverter(
    BaseDictToPbConverter[Union[FullQuaConfig, ControllerQuaConfig, LogicalQuaConfig], inc_qua_config_pb2.QuaConfig]
):
    def __init__(
        self, capabilities: ServerCapabilities, init_mode: bool = True, octave_already_configured: bool = False
    ):
        super().__init__(capabilities, init_mode)
        self.control_device_converter = ControlDeviceConverter(capabilities, init_mode)
        self.octave_converter = OctaveConverter(capabilities, init_mode)
        self.element_converter = ElementConverter(capabilities, init_mode)
        self.pulse_converter = PulseConverter(capabilities, init_mode)
        self.waveform_converter = WaveformConverter(capabilities, init_mode)
        self.iw_converter = IntegrationWeightsConverter(capabilities, init_mode)
        self.mixer_correction_converter = MixerCorrectionConverter(capabilities, init_mode)
        self.oscillator_converter = OscillatorConverter(capabilities, init_mode)
        self.octave_already_configured = octave_already_configured

    def convert(
        self,
        input_data: Union[FullQuaConfig, ControllerQuaConfig, LogicalQuaConfig],
    ) -> inc_qua_config_pb2.QuaConfig:
        self.run_preload_validations(input_data, self.octave_already_configured)

        pb_config = self.set_config_wrapper()
        controller_config = get_controller_pb_config(pb_config)
        logical_config = get_logical_pb_config(pb_config)

        def set_controllers() -> None:
            for k, v in input_data["controllers"].items():  # type: ignore[typeddict-item]
                controller_config.controlDevices[k].CopyFrom(self.control_device_converter.convert(v))
            # Controllers attribute is supported only in config v1
            if self.all_controllers_are_opx(controller_config.controlDevices) and isinstance(
                controller_config, inc_qua_config_pb2.QuaConfig.QuaConfigV1
            ):
                for _k, _v in controller_config.controlDevices.items():
                    controller_inst = get_fem_config_instance(_v.fems[OPX_FEM_IDX])
                    if not isinstance(controller_inst, inc_qua_config_pb2.QuaConfig.ControllerDec):
                        raise ValueError("This should not happen")
                    controller_config.controllers[_k].CopyFrom(controller_inst)

        def set_octaves() -> None:
            for k, v in input_data.get("octaves", {}).items():  # type: ignore[attr-defined]
                controller_config.octaves[k].CopyFrom(self.octave_converter.convert(v))

        def set_elements() -> None:
            for k, v in input_data["elements"].items():  # type: ignore[typeddict-item]
                try:
                    logical_config.elements[k].CopyFrom(self.element_converter.convert(v))
                except StickyElementIsNotSupported:
                    raise ConfigValidationException(f"Server does not support digital sticky used in element " f"'{k}'")

        def set_pulses() -> None:
            for k, v in input_data["pulses"].items():  # type: ignore[typeddict-item]
                logical_config.pulses[k].CopyFrom(self.pulse_converter.convert(v))

        def set_waveforms() -> None:
            for k, v in input_data["waveforms"].items():  # type: ignore[typeddict-item]
                logical_config.waveforms[k].CopyFrom(self.waveform_converter.convert(v))

        def set_digital_waveforms() -> None:
            for k, v in input_data["digital_waveforms"].items():  # type: ignore[typeddict-item]
                logical_config.digitalWaveforms[k].CopyFrom(
                    inc_qua_config_pb2.QuaConfig.DigitalWaveformDec(
                        samples=[
                            inc_qua_config_pb2.QuaConfig.DigitalWaveformSample(value=bool(s[0]), length=s[1])
                            for s in v["samples"]
                        ]
                    )
                )

        def set_integration_weights() -> None:
            for k, v in input_data["integration_weights"].items():  # type: ignore[typeddict-item]
                logical_config.integrationWeights[k].CopyFrom(self.iw_converter.convert(v))

        def set_mixers() -> None:
            for k, v in input_data["mixers"].items():  # type: ignore[typeddict-item]
                controller_config.mixers[k].CopyFrom(
                    inc_qua_config_pb2.QuaConfig.MixerDec(
                        correction=[self.mixer_correction_converter.convert(u) for u in v]
                    )
                )

        def set_oscillators() -> None:
            for k, v in input_data["oscillators"].items():  # type: ignore[typeddict-item]
                logical_config.oscillators[k].CopyFrom(self.oscillator_converter.convert(v))

        key_to_action = {
            "version": lambda: None,
            "controllers": set_controllers,
            "elements": set_elements,
            "pulses": set_pulses,
            "waveforms": set_waveforms,
            "digital_waveforms": set_digital_waveforms,
            "integration_weights": set_integration_weights,
            "mixers": set_mixers,
            "oscillators": set_oscillators,
            "octaves": set_octaves,
        }

        if "version" in input_data:
            warnings.warn(
                deprecation_message("version", "1.2.2", "2.0.0", "Please remove it from the QUA config."),
                DeprecationWarning,
            )

        for key in input_data:
            key_to_action[key]()

        self.apply_post_load_setters(pb_config)

        return pb_config

    def run_preload_validations(
        self,
        config: Union[FullQuaConfig, ControllerQuaConfig, LogicalQuaConfig],
        octave_already_configured: bool = False,
    ) -> None:
        if not self._init_mode:
            # With these two validations, we ensure any configuration that relates to Octave is done in init mode.
            # Or in other words, Octave doesn't support 'send program with config'.

            if "octaves" in config:
                raise OctaveUnsupportedOnUpdate("Octaves are not supported in non-init mode")

            logical_config_present = any(key in config for key in LogicalQuaConfig.__annotations__)
            if octave_already_configured and logical_config_present:
                # If Octaves were already configured, we cannot change the logical configuration anymore, because it may override
                # automatic configurations that were done for Octave, like the ones in "apply_post_load_setters()".
                raise ConfigurationLockedByOctave(
                    "Since Octaves were used in the initial configuration, no modifications to the logical configuration are allowed — whether related to Octaves or not. "
                    "To resolve this, either avoid using Octaves, or ensure all logical configuration is completed when opening the QM."
                )

    def set_config_wrapper(self) -> inc_qua_config_pb2.QuaConfig:
        pb_config = inc_qua_config_pb2.QuaConfig()

        if self._capabilities.supports(QopCaps.config_v2):
            pb_config.v2.CopyFrom(
                inc_qua_config_pb2.QuaConfig.QuaConfigV2(
                    controller_config=inc_qua_config_pb2.QuaConfig.ControllerConfig(),
                    logical_config=inc_qua_config_pb2.QuaConfig.LogicalConfig(),
                )
            )
        else:
            pb_config.v1beta.CopyFrom(inc_qua_config_pb2.QuaConfig.QuaConfigV1())

        return pb_config

    @staticmethod
    def all_controllers_are_opx(control_devices: MutableMapping[str, inc_qua_config_pb2.QuaConfig.DeviceDec]) -> bool:
        for device_config in control_devices.values():
            for fem_config in device_config.fems.values():
                _, controller_inst = which_one_of(fem_config, "fem_type_one_of")
                if not isinstance(controller_inst, inc_qua_config_pb2.QuaConfig.ControllerDec):
                    return False
        return True

    def apply_post_load_setters(self, pb_config: inc_qua_config_pb2.QuaConfig) -> None:
        # In config_v2, elements can be defined independently of mixers.
        # This breaks the existing logic, which automatically assigns default mixers and octave values based on the elements.
        # As a result, users of config_v2 that rely on these autocompleted values need to manually add them, or use the old flow (sending full config in init mode).
        # The long-term goal is to move autocompletion logic into the gateway. For more details, see: https://quantum-machines.atlassian.net/browse/OPXK-25086
        if self._init_mode:
            self.set_octave_upconverter_connection_to_elements(pb_config)
            self.set_lo_frequency_to_mix_input_elements_that_are_connected_to_octave(pb_config)
            self.set_octave_downconverter_connection_to_elements(pb_config)
            self.set_non_existing_mixers_in_mix_input_elements(pb_config)

    @staticmethod
    def set_octave_upconverter_connection_to_elements(pb_config: inc_qua_config_pb2.QuaConfig) -> None:
        octaves_config = get_controller_pb_config(pb_config).octaves
        elements_config = get_logical_pb_config(pb_config).elements

        for element_name, element in elements_config.items():
            for rf_input in element.RFInputs.values():
                device = octaves_config.get(rf_input.device_name)
                if not device:
                    raise NoMatchingOctavePort(
                        f"Octave `{rf_input.device_name}` that is connected to the element `{element_name}` (input) was not found"
                    )
                upconverter_config = device.rf_outputs.get(rf_input.port)
                if not upconverter_config:
                    raise NoMatchingOctavePort(
                        f"Port {rf_input.port} of octave `{rf_input.device_name}` that is connected to the element `{element_name}` (input) was not found"
                    )

                _, element_input = which_one_of(element, "element_inputs_one_of")
                if element_input is not None:
                    raise ElementInputConnectionAmbiguity("Ambiguous definition of element input")

                element.mixInputs.CopyFrom(
                    inc_qua_config_pb2.QuaConfig.MixInputs(
                        I=upconverter_config.I_connection, Q=upconverter_config.Q_connection
                    )
                )

    def set_lo_frequency_to_mix_input_elements_that_are_connected_to_octave(
        self, pb_config: inc_qua_config_pb2.QuaConfig
    ) -> None:
        octaves_config = get_controller_pb_config(pb_config).octaves
        elements_config = get_logical_pb_config(pb_config).elements

        for element in elements_config.values():
            _, element_input = which_one_of(element, "element_inputs_one_of")
            if isinstance(element_input, inc_qua_config_pb2.QuaConfig.MixInputs):
                rf_output = self._get_rf_output_for_octave(element, octaves_config)
                if rf_output is None:
                    continue

                if element_input.loFrequency not in {0, int(rf_output.LO_frequency)}:
                    raise ConfigValidationException(
                        "LO frequency mismatch. The frequency stated in the element is different from "
                        "the one stated in the Octave, remove the one in the element."
                    )
                element_input.loFrequency = int(rf_output.LO_frequency)
                if self._capabilities.supports_double_frequency:
                    element_input.loFrequencyDouble = rf_output.LO_frequency

    @staticmethod
    def _get_rf_output_for_octave(
        element: inc_qua_config_pb2.QuaConfig.ElementDec,
        octaves: MutableMapping[str, inc_qua_config_pb2.QuaConfig.Octave.Config],
    ) -> Optional[inc_qua_config_pb2.QuaConfig.Octave.RFOutputConfig]:
        if element.RFInputs:
            element_rf_input = list(element.RFInputs.values())[0]
            octave_config = octaves[element_rf_input.device_name]
            return octave_config.rf_outputs[element_rf_input.port]

        # This part is for users that do not use the rf_inputs  for connecting the octave
        element_input = element.mixInputs
        for octave in octaves.values():
            for rf_output in octave.rf_outputs.values():
                if all(
                    [
                        (rf_output.I_connection.controller == element_input.I.controller),
                        (rf_output.I_connection.fem == element_input.I.fem),
                        (rf_output.I_connection.number == element_input.I.number),
                        (rf_output.Q_connection.controller == element_input.Q.controller),
                        (rf_output.Q_connection.fem == element_input.Q.fem),
                        (rf_output.Q_connection.number == element_input.Q.number),
                    ]
                ):
                    return rf_output
        return None

    @staticmethod
    def set_octave_downconverter_connection_to_elements(pb_config: inc_qua_config_pb2.QuaConfig) -> None:
        octaves_config = get_controller_pb_config(pb_config).octaves
        elements_config = get_logical_pb_config(pb_config).elements

        for element_name, element in elements_config.items():
            for _, rf_output in element.RFOutputs.items():
                device = octaves_config.get(rf_output.device_name)
                if not device:
                    raise NoMatchingOctavePort(
                        f"Octave `{rf_output.device_name}` that is connected to the element `{element_name}` (output) was not found"
                    )
                port = device.rf_inputs.get(rf_output.port)
                if not port:
                    raise NoMatchingOctavePort(
                        f"Port {rf_output.port} of octave `{rf_output.device_name}` that is connected to the element `{element_name}` (output) was not found"
                    )

                downconverter_config = device.if_outputs
                outputs_form_octave = {
                    downconverter_config.IF_out1.name: downconverter_config.IF_out1.port,
                    downconverter_config.IF_out2.name: downconverter_config.IF_out2.port,
                }
                for k, v in outputs_form_octave.items():
                    if k in element.outputs:
                        if v != element.outputs[k]:
                            raise ElementOutputConnectionAmbiguity(
                                f"Output {k} is connected to {element.outputs[k]} but the octave "
                                f"downconverter is connected to {v}"
                            )
                    else:
                        element.outputs[k].CopyFrom(v)
                        _, element_outputs = which_one_of(element, "element_outputs_one_of")
                        if isinstance(element_outputs, inc_qua_config_pb2.QuaConfig.MicrowaveOutputPortReference):
                            raise ConfigValidationException("Cannot connect octave to microwave output")
                        elif isinstance(element_outputs, inc_qua_config_pb2.QuaConfig.MultipleOutputs):
                            element_outputs.port_references[k].CopyFrom(v)
                        else:
                            element.multipleOutputs.CopyFrom(
                                inc_qua_config_pb2.QuaConfig.MultipleOutputs(port_references={k: v})
                            )

    @staticmethod
    def set_non_existing_mixers_in_mix_input_elements(pb_config: inc_qua_config_pb2.QuaConfig) -> None:
        mixers_config = get_controller_pb_config(pb_config).mixers
        elements_config = get_logical_pb_config(pb_config).elements

        for element_name, element in elements_config.items():
            _, element_input = which_one_of(element, "element_inputs_one_of")
            if isinstance(element_input, inc_qua_config_pb2.QuaConfig.MixInputs):
                if element.HasField(
                    "intermediateFrequency"
                ):  # This is here because in validation I saw that we can set an element without IF
                    if not element_input.mixer:
                        element_input.mixer = f"{element_name}_mixer_{uuid.uuid4().hex[:3]}"
                        # The uuid is just to make sure the mixer doesn't exist
                    if element_input.mixer not in mixers_config:
                        mixers_config[element_input.mixer].CopyFrom(
                            inc_qua_config_pb2.QuaConfig.MixerDec(
                                correction=[
                                    inc_qua_config_pb2.QuaConfig.CorrectionEntry(
                                        frequency=element.intermediateFrequency.value,
                                        frequencyNegative=element.intermediateFrequencyNegative,
                                        frequencyDouble=element.intermediateFrequencyDouble,
                                        loFrequency=element_input.loFrequency,
                                        loFrequencyDouble=element_input.loFrequencyDouble,
                                        correction=inc_qua_config_pb2.QuaConfig.Matrix(v00=1, v01=0, v10=0, v11=1),
                                    )
                                ]
                            )
                        )

    def deconvert(self, output_data: inc_qua_config_pb2.QuaConfig) -> FullQuaConfig:
        controller_config = get_controller_pb_config(output_data)
        logical_config = get_logical_pb_config(output_data)

        if controller_config.controlDevices:
            controllers = {
                name: self.control_device_converter.deconvert(value)
                for name, value in controller_config.controlDevices.items()
            }
        elif isinstance(controller_config, inc_qua_config_pb2.QuaConfig.QuaConfigV1) and controller_config.controllers:
            controllers = {
                name: self.control_device_converter._deconvert_controller(value)
                for name, value in controller_config.controllers.items()
            }
        else:
            controllers = {}

        result: FullQuaConfig = {
            "controllers": controllers,
            "oscillators": {
                name: self.oscillator_converter.deconvert(oscillator)
                for name, oscillator in logical_config.oscillators.items()
            },
            "elements": {
                name: self.element_converter.deconvert(elem) for name, elem in logical_config.elements.items()
            },
            "pulses": {name: self.pulse_converter.deconvert(pulse) for name, pulse in logical_config.pulses.items()},
            "waveforms": {name: self.waveform_converter.deconvert(wf) for name, wf in logical_config.waveforms.items()},
            "digital_waveforms": {
                name: self._deconvert_digital_waveforms(wf) for name, wf in logical_config.digitalWaveforms.items()
            },
            "integration_weights": {
                name: self.iw_converter.deconvert(iw) for name, iw in logical_config.integrationWeights.items()
            },
            "mixers": {
                name: [self.mixer_correction_converter.deconvert(u) for u in value.correction]
                for name, value in controller_config.mixers.items()
            },
        }

        return result

    @staticmethod
    def _deconvert_digital_waveforms(
        digital_wave_form: inc_qua_config_pb2.QuaConfig.DigitalWaveformDec,
    ) -> DigitalWaveformConfigType:
        temp_list: list[tuple[int, int]] = []
        for sample in digital_wave_form.samples:
            value = int(bool(sample.value))
            temp_list.append((value, sample.length))

        return {"samples": temp_list}



import logging
from typing import Type, Tuple, Iterator

from qm.program import Program
from qm.simulate import interface
from qm.api.base_api import BaseApi
from qm.utils.protobuf_utils import LOG_LEVEL_MAP
from qm.simulate.interface import SimulationConfig
from qm.grpc.qm.pb.frontend_pb2_grpc import FrontendStub
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QMSimulationError, FailedToExecuteJobException
from qm.grpc.qm.pb import frontend_pb2, job_results_pb2, inc_qua_config_pb2
from qm.api.models.compiler import CompilerOptionArguments, get_request_compiler_options

logger = logging.getLogger(__name__)


def create_simulation_request(
    config: inc_qua_config_pb2.QuaConfig,
    program: Program,
    simulate: SimulationConfig,
    compiler_options: CompilerOptionArguments,
    capabilities: ServerCapabilities,
) -> frontend_pb2.SimulationRequest:
    if not isinstance(program, Program):
        raise Exception("program argument must be of type qm.program.Program")

    request = frontend_pb2.SimulationRequest()

    if hasattr(config, "v1_beta"):
        config.v1_beta._unknown_fields = b""  # This correction is for a return of unknown fields from
        # an old GW, that changes the flow of the config parsing (in the GW, and to a buggy one).
        # Check a year from now (This change was done on Feb/2024) if this is still necessary.
    request.config.CopyFrom(config)

    if isinstance(simulate, SimulationConfig):
        request.simulate.CopyFrom(
            frontend_pb2.ExecutionRequest.Simulate(
                duration=simulate.duration,
                includeAnalogWaveforms=simulate.include_analog_waveforms,
                includeDigitalWaveforms=simulate.include_digital_waveforms,
                extraProcessingTimeoutMs=simulate.extraProcessingTimeoutInMs,
            )
        )
        request = simulate.update_simulate_request(request, capabilities)

        for connection in simulate.controller_connections:
            if not isinstance(connection.source, type(connection.target)):
                raise Exception(
                    f"Unsupported InterOpx connection. Source is "
                    f"{type(connection.source).__name__} but target is "
                    f"{type(connection.target).__name__}"
                )

            if isinstance(connection.source, interface.InterOpxAddress) and isinstance(
                connection.target, interface.InterOpxAddress
            ):
                con = frontend_pb2.InterOpxConnection(
                    addressToAddress=frontend_pb2.InterOpxConnection.AddressToAddress(
                        source=frontend_pb2.InterOpxAddress(
                            controller=connection.source.controller,
                            left=connection.source.is_left_connection,
                        ),
                        target=frontend_pb2.InterOpxAddress(
                            controller=connection.target.controller,
                            left=connection.target.is_left_connection,
                        ),
                    )
                )
            elif isinstance(connection.source, interface.InterOpxChannel) and isinstance(
                connection.target, interface.InterOpxChannel
            ):
                con = frontend_pb2.InterOpxConnection(
                    channelToChannel=frontend_pb2.InterOpxConnection.ChannelToChannel(
                        source=frontend_pb2.InterOpxChannel(
                            controller=connection.source.controller,
                            channelNumber=connection.source.channel_number,
                        ),
                        target=frontend_pb2.InterOpxChannel(
                            controller=connection.target.controller,
                            channelNumber=connection.target.channel_number,
                        ),
                    )
                )
            else:
                raise Exception(
                    f"Unsupported InterOpx connection. Source is "
                    f"{type(connection.source).__name__}. Supported types are "
                    f"frontend_pb2.InterOpxAddress "
                    f"or frontend_pb2.InterOpxChannel"
                )

            request.controllerConnections.append(con)

    request.highLevelProgram.CopyFrom(program.qua_program)
    request.highLevelProgram.compilerOptions.CopyFrom(get_request_compiler_options(compiler_options))
    return request


class SimulationApi(BaseApi[FrontendStub]):
    def __init__(self, connection_details: ConnectionDetails):
        super().__init__(connection_details)
        self._timeout = None

    @property
    def _stub_class(self) -> Type[FrontendStub]:
        return FrontendStub

    def simulate(
        self,
        config: inc_qua_config_pb2.QuaConfig,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: CompilerOptionArguments,
        capabilities: ServerCapabilities,
    ) -> Tuple[str, frontend_pb2.SimulatedResponsePart]:
        if type(program) is not Program:
            raise Exception("program argument must be of type qm.program.Program")

        request = create_simulation_request(config, program, simulate, compiler_options, capabilities)
        logger.info("Simulating program")

        response = self._run(self._stub.Simulate, request, timeout=self._timeout)

        messages = [(LOG_LEVEL_MAP[msg.level], msg.message) for msg in response.messages]

        config_messages = [(LOG_LEVEL_MAP[msg.level], msg.message) for msg in response.configValidationErrors]

        job_id = response.jobId

        for lvl, msg in messages:
            logger.log(lvl, msg)

        for lvl, msg in config_messages:
            logger.log(lvl, msg)

        if not response.success:
            logger.error("Job " + job_id + " failed. Failed to execute program.")
            for error in response.simulated.errors:
                logger.error(f"Simulation error: {error}")
            raise FailedToExecuteJobException(job_id)

        return job_id, response.simulated

    def get_simulated_quantum_state(self, job_id: str) -> frontend_pb2.DensityMatrix:
        request = frontend_pb2.GetSimulatedQuantumStateRequest(jobId=job_id)
        response: frontend_pb2.GetSimulatedQuantumStateResponse = self._run(
            self._stub.GetSimulatedQuantumState, request, timeout=self._timeout
        )

        if response.ok:
            return response.state

        raise QMSimulationError("Error while pulling quantum state")

    def pull_simulator_samples(
        self, job_id: str, include_analog: bool, include_digital: bool
    ) -> Iterator[job_results_pb2.SimulatorSamplesResponse]:
        request = job_results_pb2.PullSimulatorSamplesRequest(
            jobId=job_id,
            includeAnalog=include_analog,
            includeDigital=include_digital,
            includeAllConnections=True,  # TODO: Check whether it should appear
        )
        return self._run_iterator(self._stub.PullSimulatorSamples, request)



from qm.grpc.qm.pb import inc_qua_pb2 as _inc_qua_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from qm.grpc.qm.pb import qm_manager_pb2 as _qm_manager_pb2
from qm.grpc.qm.pb import general_messages_pb2 as _general_messages_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class QuaValues(_message.Message):
    __slots__ = ("int_value", "double_value", "boolean_value")
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
    int_value: int
    double_value: float
    boolean_value: bool
    def __init__(self, int_value: _Optional[int] = ..., double_value: _Optional[float] = ..., boolean_value: bool = ...) -> None: ...

class CompileRequest(_message.Message):
    __slots__ = ("program", "jobId")
    PROGRAM_FIELD_NUMBER: _ClassVar[int]
    JOBID_FIELD_NUMBER: _ClassVar[int]
    program: _inc_qua_pb2.QuaProgram
    jobId: str
    def __init__(self, program: _Optional[_Union[_inc_qua_pb2.QuaProgram, _Mapping]] = ..., jobId: _Optional[str] = ...) -> None: ...

class CompileResponse(_message.Message):
    __slots__ = ("ok", "messages", "binary", "metadata", "debug")
    class DebugEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    OK_FIELD_NUMBER: _ClassVar[int]
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    BINARY_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    DEBUG_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    messages: _containers.RepeatedCompositeFieldContainer[CompilerMessage]
    binary: bytes
    metadata: str
    debug: _containers.ScalarMap[str, str]
    def __init__(self, ok: bool = ..., messages: _Optional[_Iterable[_Union[CompilerMessage, _Mapping]]] = ..., binary: _Optional[bytes] = ..., metadata: _Optional[str] = ..., debug: _Optional[_Mapping[str, str]] = ...) -> None: ...

class CompilerMessage(_message.Message):
    __slots__ = ("message", "level")
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    message: str
    level: _general_messages_pb2.MessageLevel
    def __init__(self, message: _Optional[str] = ..., level: _Optional[_Union[_general_messages_pb2.MessageLevel, str]] = ...) -> None: ...

class ValidationResponse(_message.Message):
    __slots__ = ("messages",)
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    messages: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.ConfigValidationMessage]
    def __init__(self, messages: _Optional[_Iterable[_Union[_qm_manager_pb2.ConfigValidationMessage, _Mapping]]] = ...) -> None: ...

class DynamicConfig(_message.Message):
    __slots__ = ("version", "root")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    ROOT_FIELD_NUMBER: _ClassVar[int]
    version: int
    root: _struct_pb2.Struct
    def __init__(self, version: _Optional[int] = ..., root: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...
